#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <math.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"


#ifdef __GLIBC__
#include <sys/io.h>
#endif

#include <fcntl.h>

// below is logic needed to set laptop bidirectional port to output
#define DIRBIT 0x20

extern  GtkWidget *app1;

// new port defines to get 0x378 working on lp0
#define LPTBASE 0x378
// possible alternates if you have weird hardware
//#define LPTBASE 0x278
//#define LPTBASE 0x3bc

void decoderxch(int ch);
void decodetxch(int ch);
void decoderxcg( int ch );
void decodetxcg( int ch );
void decoderom(void);
const char *decodecg( int code );
void decodecct(int ch);
void read_filename (GtkWidget *widget, gpointer user_data);
void encoderom( void );


static const unsigned short LPTDAT = LPTBASE;
static const unsigned short LPTSTAT = LPTBASE + 1;
static const unsigned short LPTCTRL = LPTBASE + 2;
static double ref;          /* reference oscillator freq in hz */
static double space;        /* channel spacing in hz */

static int fd0 = -1;
const gchar *selected_filename;

static unsigned char x2212[256];    /* binary earom image */

/* these are arrays of similar field names so we can look up the */
/* widgets from inside a loop                                    */
char *rxfreq[16] = {"ch01rx", "ch02rx", "ch03rx", "ch04rx",
                    "ch05rx", "ch06rx", "ch07rx", "ch08rx",
                    "ch09rx", "ch10rx", "ch11rx", "ch12rx",
                    "ch13rx", "ch14rx", "ch15rx", "ch16rx" };

char *rxcg[16] =   {"ch01rxcg", "ch02rxcg", "ch03rxcg", "ch04rxcg",
                    "ch05rxcg", "ch06rxcg", "ch07rxcg", "ch08rxcg",
                    "ch09rxcg", "ch10rxcg", "ch11rxcg", "ch12rxcg",
                    "ch13rxcg", "ch14rxcg", "ch15rxcg", "ch16rxcg" };

char *txfreq[16] = {"ch01tx", "ch02tx", "ch03tx", "ch04tx",
                    "ch05tx", "ch06tx", "ch07tx", "ch08tx",
                    "ch09tx", "ch10tx", "ch11tx", "ch12tx",
                    "ch13tx", "ch14tx", "ch15tx", "ch16tx" };

char *txcg[16] =   {"ch01txcg", "ch02txcg", "ch03txcg", "ch04txcg",
                    "ch05txcg", "ch06txcg", "ch07txcg", "ch08txcg",
                    "ch09txcg", "ch10txcg", "ch11txcg", "ch12txcg",
                    "ch13txcg", "ch14txcg", "ch15txcg", "ch16txcg" };

char *cct[16] =    {"ch01cct", "ch02cct", "ch03cct", "ch04cct",
                    "ch05cct", "ch06cct", "ch07cct", "ch08cct",
                    "ch09cct", "ch10cct", "ch11cct", "ch12cct",
                    "ch13cct", "ch14cct", "ch15cct", "ch16cct" };
                    
/* convert 2 ascii chars in hex to a decimal number */
static int hex2( unsigned char a, unsigned char b)
{
    unsigned int aa,bb;
    unsigned int result;

    aa = tolower(a);
    if ( aa >= 'a' && aa <= 'f' )
        aa = aa - 'a' + 10;
    else
        aa = aa - '0';


    bb = tolower(b);
    if ( bb >= 'a' && bb <= 'f' )
        bb = bb - 'a' + 10;
    else
        bb = bb - '0';

    result = aa*16 + bb;    
    
    return result;
}



//=====================================================================
// 
// 
//=====================================================================
static void done()
{
    /* release parallel port */
    if (fd0 > 0)
        close(fd0);
    ioperm(LPTDAT, 3, 0);
}
//=====================================================================
// 
// 
//=====================================================================
const char *decodecg( int code )
{
    static char buff[256];

    /* bit 0x08 at offset 7 is the squelch tail eliminate bit (used on rx) */
    
    switch (code & 0xf3)
    {
    case 0x00: strcpy(buff,"No CG"); break;
    case 0x10: strcpy(buff,"67.0 "); break;
    case 0x20: strcpy(buff,"71.9 "); break;
    case 0x30: strcpy(buff,"74.4 "); break;
    case 0x40: strcpy(buff,"77.0 "); break;    
    case 0x50: strcpy(buff,"79.7 "); break;
    case 0x60: strcpy(buff,"82.5 "); break;
    case 0x70: strcpy(buff,"85.4 "); break;
    case 0x80: strcpy(buff,"88.5 "); break;
    case 0x90: strcpy(buff,"91.5 "); break;
    case 0xa0: strcpy(buff,"94.8 "); break;
    case 0x12: strcpy(buff,"97.4 "); break;
    case 0xb0: strcpy(buff,"100.0"); break;
    case 0xc0: strcpy(buff,"103.5"); break;
    case 0xd0: strcpy(buff,"107.2"); break;
    case 0xe0: strcpy(buff,"110.9"); break;
    case 0xf0: strcpy(buff,"114.8"); break;
    case 0x01: strcpy(buff,"118.8"); break;
    case 0x11: strcpy(buff,"123.0"); break;
    case 0x21: strcpy(buff,"127.3"); break;
    case 0x31: strcpy(buff,"131.8"); break;
    case 0x41: strcpy(buff,"136.5"); break;
    case 0x51: strcpy(buff,"141.3"); break;
    case 0x61: strcpy(buff,"146.2"); break;
    case 0x71: strcpy(buff,"151.4"); break;
    case 0x81: strcpy(buff,"156.7"); break;
    case 0x91: strcpy(buff,"162.2"); break;
    case 0xa1: strcpy(buff,"167.9"); break;
    case 0xb1: strcpy(buff,"173.8"); break;
    case 0xc1: strcpy(buff,"179.9"); break;
    case 0xd1: strcpy(buff,"186.2"); break;
    case 0xe1: strcpy(buff,"192.8"); break;    
    case 0xf1: strcpy(buff,"203.5"); break;    
    case 0x02: strcpy(buff,"210.7"); break;
    default:   strcpy(buff,"error");
    }
    if (code & 0x08)
        strcat(buff," ste");
    return( buff );
}
//=====================================================================
// 
// 
//=====================================================================
/* compute and print rx freq from the prom contents for ch 1-16*/
void decoderxch(int ch)
{
    int r,divider;      /* divider */
    unsigned char *chbuff;
    unsigned char buff[20];
    GtkWidget *wid;
    
    if ( ch >= 1 && ch <= 16 )
        wid = lookup_widget(app1,rxfreq[ch-1]);
    else
    {
        fprintf(stderr,"bad call to decoderxch %d\n",ch); 
        return;
    }
    
    
    /* point to the 8 bytes of rx channel data for specified channel */
    chbuff = &x2212[0] + ((ch % 16) * 16);
    /* there are 20 packed bits of freq info... lets build the numbers */
    if (chbuff[5] == 0x0f)
    {
        gtk_entry_set_text((GtkEntry *)wid, "000.0000");
    }
    else
    {
        r = chbuff[1] & 0x07;
        if ((r & 0x04) == 0)
            ref = 13.2e6;
        else
            ref = 13.8e6; 
        
        space = 5000.0;
        switch ( r & 0x03 )
        {
        case 0x01: space = 12500.0/3.0; break;
        case 0x02: space = 5000.0;      break;
        case 0x03: space = 6250.0;      break;
        default: fprintf(stderr,"Invalid channel spacing code %d\n",r);
        }

//      g_print("ref = %fhz, spacing = %fhz \n",ref,space);        
        
        /* build a and n divder by unpacking bits lsb to msb */
        divider =(chbuff[4] & 0x0f) +         /* a3.a2.a1.a0 */
                ((chbuff[5] & 0x03) << 4) +   /* a5,a4 */
                ((chbuff[5] & 0x08) << 3) +   /* n0 */
                ((chbuff[2] & 0x0f) << 7) +   /* n4,n3,n2,n1 */
                ((chbuff[3] & 0x0f) << 11) +  /* n8,n7,n6,n5 */
                ((chbuff[1] & 0x08) << 12);   /* n9 */
        sprintf( buff,"%8.4f", (divider * space - 45.0e6)/1.0e6 );
        gtk_entry_set_text((GtkEntry *)wid, buff);        
    }
}
//=====================================================================
// 
// 
//=====================================================================
/* decode and print the tx channel freq.... assumes rx for ch has already */
/* been printed  (determines ch spacing and ref divider)                  */
void decodetxch(int ch)
{
    int divider;      /* divider */
    unsigned char *chbuff;
    unsigned char buff[20];
    GtkWidget *wid;
    
    if ( ch >= 1 && ch <= 16 )
        wid = lookup_widget(app1,txfreq[ch-1]);
    else
    {
        fprintf(stderr,"bad call to decodetxch %d\n",ch); 
        return;
    }


    /* point to the 8 bytes of tx channel data for specified channel */
    chbuff = &x2212[0] + ((ch % 16) * 16) + 8;
    /* there are 20 packed bits of freq info... lets build the numbers */
    if (chbuff[5] == 0x0f)
    {
        gtk_entry_set_text((GtkEntry *)wid, "000.0000");
    }
    else
    {
        /* build a and n divder by unpacking bits lsb to msb */
        divider =(chbuff[4] & 0x0f) +         /* a3.a2.a1.a0 */
                ((chbuff[5] & 0x03) << 4) +   /* a5,a4 */
                ((chbuff[5] & 0x08) << 3) +   /* n0 */
                ((chbuff[2] & 0x0f) << 7) +   /* n4,n3,n2,n1 */
                ((chbuff[3] & 0x0f) << 11) +  /* n8,n7,n6,n5 */
                ((chbuff[1] & 0x08) << 12);   /* n9 */
        sprintf( buff,"%8.4f", (divider * space)/1.0e6 );
        gtk_entry_set_text((GtkEntry *)wid, buff);        
    }
}
//=====================================================================
// 
// 
//=====================================================================
/* decode the carrier timer enable bit for the tx channel */
void decodecct(int ch)
{
    GtkWidget *wid;
    unsigned char *chbuff;
        
    if ( ch >= 1 && ch <= 16 )
        wid = lookup_widget(app1,cct[ch-1]);
    else
    {
        fprintf(stderr,"bad call to decodecct %d\n",ch); 
        return;
    }


    /* point to the 8 bytes of tx channel data for specified channel */
    chbuff = &x2212[0] + ((ch % 16) * 16) + 8;
    /* there are 20 packed bits of freq info... lets build the numbers */
    if ((chbuff[1] & 0x04) == 0x04)
        gtk_toggle_button_set_active((GtkToggleButton *)wid, TRUE);
    else
        gtk_toggle_button_set_active((GtkToggleButton *)wid, FALSE);

}
//=====================================================================
// 
// 
//=====================================================================
void decoderxcg( int ch )
{
    unsigned char *chbuff;
    unsigned int code;
    unsigned char buff[20];
    GtkWidget *wid;
    
    if ( ch >= 1 && ch <= 16 )
        wid = lookup_widget(app1,rxcg[ch-1]);
    else
    {
        fprintf(stderr,"bad call to decoderxcg %d\n",ch); 
        return;
    }


    /* point to the 2 bytes of rx cg data for specified channel */
    chbuff = &x2212[0] + ((ch % 16) * 16) + 6;

    code = (((chbuff[0] & 0x0f) << 4) + (chbuff[1] & 0x0f));
    sprintf(buff,"%02X = %s",code,decodecg( code ));
    gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), buff);            
}
//=====================================================================
// 
// 
//=====================================================================
void decodetxcg(int ch)
{
    unsigned char *chbuff;
    unsigned int code;
    unsigned char buff[20];
    GtkWidget *wid;
    
    if ( ch >= 1 && ch <= 16 )
        wid = lookup_widget(app1,txcg[ch-1]);
    else
    {
        fprintf(stderr,"bad call to decodetxcg %d\n",ch);
        return;
    }
    

    if ( wid == NULL )
        fprintf(stderr,"Opps, lookup of chxxtxcg widget failed\n");    
    /* point to the 8 bytes of tx channel data for specified channel */
    chbuff = &x2212[0] + ((ch % 16) * 16) + 0x0e;
    code = (((chbuff[0] & 0x0f) << 4) + (chbuff[1] & 0x0f));

    sprintf(buff,"%02X = %s",code,decodecg( code ));
    gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), buff);        
}
//=====================================================================
// 
// 
//=====================================================================
/* this routine attempts to decode and print the rom contents */
/* in a channel by channel format                             */
void decoderom(void)
{
    int ch;
    unsigned char buff[80];     // this was too small at 20!!!
    GtkWidget *wid;
    
    wid = lookup_widget(app1,"appbar1");
   	gnome_appbar_set_status((GnomeAppBar *)wid,"Decoding EPROM Contents");    
    /* print out special locations in rom */
    wid = lookup_widget(app1,"numch");
    sprintf( buff,"Number of Active Channels: %d",(x2212[0x80]&0x0f) + 1);
   	gtk_label_set_text((GtkLabel *)wid,buff);
    wid = lookup_widget(app1,"txtimer");

//gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), buff);
    switch( x2212[0xf8]&0x0f )
    {
    case 0: gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "0 = 30 seconds");break;
    case 1: gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "1 = 1 minute");break;
    case 2: gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "2 = 1.5 minutes");break;
    case 3: gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "3 = 2 minutes");break;
    case 4: gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "4 = 2.5 minutes");break;
    case 5: gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "5 = 3 minutes");break;
    case 6: gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "6 = 3.5 minutes");break;
    case 7: gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "7 = 4 minutes");break;
    default:  gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "5 = 3 minutes");   
    }

    /* scan option code */
    wid = lookup_widget(app1,"scanoption");
    switch ( x2212[0x00] & 0x0f )
    {
    case 0x03:  /* this doesnt match the doc, but seems to work */
    case 0x0f:  gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "F = Front Programmable");break;
    case 0x00:  gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "0 = Fixed Priority Channel");
                wid = lookup_widget(app1,"fixch");
                sprintf( buff,"%d",(x2212[0x40]&0x0f)+1);
                gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), buff);
                break;
    case 0x01:  gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "1 = Fixed Priority to Follow Channel");break;
    default:    gtk_entry_set_text((GtkEntry *)(((GtkCombo *)wid)->entry), "F = Front Programmable");
                fprintf(stderr,"Invalid scan option in rom image = %d\n",x2212[0x00] & 0x0f);
    }

    for (ch=1; ch<=16; ch++)
    {
        decoderxch(ch);     /* rx freq */
        decoderxcg(ch);     /* rx channel guard tone*/
        decodetxch(ch);     /* tx freq */
        decodetxcg(ch);     /* tx channel guard tone */
        decodecct(ch);      /* carrier controlled timer state */
    }
    wid = lookup_widget(app1,"appbar1");
   	gnome_appbar_set_status((GnomeAppBar *)wid,"");    

}


//=====================================================================
// 
//  create a new setup (set screen to defaults)
//=====================================================================
void
on_new1_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    int i;
    /* start with a clean image */
    memset(&x2212[0],0,sizeof(x2212));

    /* start with all ch disabled */
    for( i = 5; i<=255; i+=8)
        x2212[i] = 0x0f;    

    /* make front panel prog scan */
    x2212[0x00] = 0x0f;
    
    /* set up 1 channel in radio */
    x2212[0x80] = 0;
    
    /* set timeout timer to 3 minutes */
    x2212[0xf8] = 5;
    
    /* set fixed pri channel to 1 */
    x2212[0x40] = 0;
    
    /* set pri ch special code ??? always 0 */
    x2212[0xe0] = 0;
    
    /* set up the only rx ch as 146.520mhz */
    x2212[0x11] = 0x0a;
    x2212[0x12] = 0x0b;
    x2212[0x13] = 0x02;
    x2212[0x14] = 0x00;
    x2212[0x15] = 0x02;
    
    /* show user what we got */
    decoderom();
}

//=====================================================================
// 
//  open an existing hex file and show us the contents
//=====================================================================
void
on_open1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *file_selector;
    file_selector = gtk_file_selection_new("Hex Filename To Read...");
    gtk_file_selection_complete(GTK_FILE_SELECTION(file_selector),"*.hex");

    g_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->ok_button),
                     "clicked",
                     G_CALLBACK (read_filename),
                     (gpointer) file_selector);
   			   
   /* Ensure that the dialog box is destroyed when the user clicks a button. */
   
   g_signal_connect_swapped (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->ok_button),
                             "clicked",
                             G_CALLBACK (gtk_widget_destroy), 
                             (gpointer) file_selector); 

   g_signal_connect_swapped (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->cancel_button),
                             "clicked",
                             G_CALLBACK (gtk_widget_destroy),
                             (gpointer) file_selector); 
   
   /* Display that dialog */
   gtk_widget_show (file_selector);
}

//=====================================================================
// 
// The file selection widget and the string to read the chosen filename
//=====================================================================
void read_filename (GtkWidget *widget, gpointer user_data) 
{
    FILE *fp;
    int i;
    GtkWidget *wid;
    GtkWidget *file_selector = (GtkWidget *)user_data;
    unsigned char lb[81];            /* buffer used for reading chunks of file */
    int rc;
    int cnt;
    int offset = 0;          
    
    /* start with a clean image */
    memset(x2212,0,sizeof(x2212));

    /* start with all ch disabled */
    for( i = 5; i<=255; i+=8)
        x2212[i] = 0x0f;    

    wid = lookup_widget(app1,"appbar1");

    selected_filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_selector));
//    g_print ("Selected filename: %s\n", selected_filename);
    
    fp = fopen(selected_filename,"rt");
    if ( fp == NULL )
    {
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"ERROR - unable to open file");
    }
    else
    {
        /* the hack is is that we assume that the hex file is well behaved */
        /* and is sequencial bytes of data (we ignore the address offset)  */
        /* we also ignore the checksum byte at the end of each line        */
        while ((rc = fgetc(fp)) != EOF )
        {
            char *result;
            /* all valid lines should start with ':' */
            if ((char)rc != ':' ) continue;

            /* read 2  digit count ofhex entries on this line */
            result = fgets(&lb[0],3,fp); 
            if (isxdigit(lb[0]) && isxdigit(lb[1]))
            {
                cnt = hex2(lb[0],lb[1]);
                if( cnt == 0 ) continue;
            }
            else
            {
                cnt = 0;
                continue;
            }
        
            /* now read and skip the 4 digit offset and the 2 digit rec type */
        
            result = fgets(&lb[0],7,fp); 
            for ( i= 0; i < cnt; i++ )
            {
                /* read all the hex data bytes from line */
                result = fgets(&lb[0],3,fp); 
                if (isxdigit(lb[0]) && isxdigit(lb[1]))
                    x2212[offset++] = (unsigned char)hex2(lb[0],lb[1]);              

            }
            /* all that remains is checksum and crlf... discard em */
        }
  
        if ( fp )
            fclose(fp);
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"Read file");
    }
    /* show user what we read */
    decoderom();
}


//=====================================================================
// 
// 
//=====================================================================
/* The file selection widget and the string to store the chosen filename */

void store_filename (GtkWidget *widget, gpointer user_data) 
{
    FILE *fp;
    char cksum;
    int i;
    GtkWidget *wid;
    GtkWidget *file_selector = (GtkWidget *)user_data;
    
    wid = lookup_widget(app1,"appbar1");

    selected_filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_selector));
//    g_print ("Selected filename: %s\n", selected_filename);
    
    /* convert screen to binary image */
    encoderom();
    
    /* convert binary image to hex format */
    /* write to user specified file name */
    cksum = 0;
    
    fp = fopen(selected_filename,"wt");
    if ( fp == NULL )
    {
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"ERROR - unable to open file");
        return;
    }
    else
    {
        /* write out an mcs8 intel hex format file of the rom image */
        for ( i=0; i<256; i++)
        {
            if ( i % 16 == 0 )
            {
                fprintf(fp,":10%04X00",i);  /* 16 bytes data,addr, rectype=0 */
                cksum = (0x10 + (i>>8) + i) & 0xFF;
            }
            fprintf(fp,"%02X",x2212[i]);         /* write 1 data byte in hex */
            cksum += x2212[i];
            
            if ( i % 16 == 15 )             /* last byte on line */
            {
                fprintf(fp,"%02X\n",-cksum & 0xFF);    /* write cksum, newline */
            }       
        }
        fprintf(fp,":00000001FF\n");        /* end of file marker */
        fflush(fp);
        fclose(fp);
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"Wrote file");
    }
    /* update to make sure screen shows user what we saved */
    decoderom();
}

//=====================================================================
// 
//  save existing screen data in user selected filename
//=====================================================================
void
on_save_as1_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *file_selector;
    file_selector = gtk_file_selection_new("Hex Filename To Save...");
    gtk_file_selection_complete(GTK_FILE_SELECTION(file_selector),"*.hex");

    g_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->ok_button),
                     "clicked",
                     G_CALLBACK (store_filename),
                     (gpointer) file_selector);
   			   
   /* Ensure that the dialog box is destroyed when the user clicks a button. */
   
   g_signal_connect_swapped (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->ok_button),
                             "clicked",
                             G_CALLBACK (gtk_widget_destroy), 
                             (gpointer) file_selector); 

   g_signal_connect_swapped (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->cancel_button),
                             "clicked",
                             G_CALLBACK (gtk_widget_destroy),
                             (gpointer) file_selector); 
   
   /* Display that dialog */
   gtk_widget_show (file_selector);
}

//=====================================================================
// 
// 
//=====================================================================
void
on_quit1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    printf("on_exit1_activate\n");
    done();
	gtk_main_quit();

}

//=====================================================================
// 
// 
//=====================================================================
void
on_read1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *wid;

    wid = lookup_widget(app1,"appbar1");
   	gnome_appbar_set_status((GnomeAppBar *)wid,"Reading Rom in Adapter");
    gnome_appbar_set_progress_percentage((GnomeAppBar *)wid,0.0);  

    /* read rom image in from adapter... store in global x2212 array */
    int i;
   
    outb(4 & ~DIRBIT,LPTCTRL);

    /* do a recall to transfer eeprom image to the ram mirror */
    outb(12 & ~DIRBIT,LPTCTRL);
    usleep(1);
    outb(4 & ~DIRBIT,LPTCTRL);            
    usleep(1);
    
    for (i=0; i<256; i++)
    {
        outb(i,LPTDAT);
        usleep(1);
        gtk_main_iteration_do(FALSE);        
        outb(6 & ~DIRBIT,LPTCTRL);
        usleep(1);
        gtk_main_iteration_do(FALSE);        
        x2212[i] = (inb(LPTSTAT)/8) & 0x0f;
        outb(4 & ~DIRBIT,LPTCTRL);
        gnome_appbar_set_progress_percentage((GnomeAppBar *)wid,(double)i/255.0);  
        gtk_main_iteration_do(FALSE);
    }
    gnome_appbar_set_progress_percentage((GnomeAppBar *)wid,0.0);  
   	gnome_appbar_set_status((GnomeAppBar *)wid,"");    
    decoderom();

}

//=====================================================================
// 
// hex2bin : convert a text character to a binary number 0..15
//=====================================================================
unsigned char hex2bin( unsigned char in )
{
    if ( in >= '0' && in <= '9' )
        return( in - '0' );
        
    if ( in >= 'A' && in <= 'F' )
        return( in - 55 );

    if ( in >= 'a' && in <= 'f' )
        return( in - 87 );

    fprintf(stderr,"Bad hex character conversion >>%c<< = %02X\n",in,in);
    return 0;
}


//=====================================================================
// 
// Build rom image from what is on the screen
//=====================================================================
void encoderom( void )
{
    int i;
    int chan; 
    double freq, realfreq;
    int divisor;
    unsigned char *chbuff;
    
    GtkWidget *wid;
    G_CONST_RETURN gchar *txt;
    
    /* start with a clean image */
    memset(x2212,0,sizeof(x2212));

    /* start with all ch disabled */
    for( i = 5; i<=255; i+=8)
        x2212[i] = 0x0f;    
    /* set special fields */
    wid = lookup_widget(app1,"scanoption");
    txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));
    x2212[0x00] = hex2bin(txt[0]);

    wid = lookup_widget(app1,"txtimer");
    txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));
    x2212[0xf8] = hex2bin(txt[0]);

    wid = lookup_widget(app1,"fixch");
    txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));
    if ( txt[1] != 0 )
    {
        /* two digit conversion */
        x2212[0x40] = ((hex2bin( txt[0] ) * 10) + hex2bin(txt[1])) - 1;
    }
    else
        /* single digit conversion */
        x2212[0x40] = hex2bin(txt[0]) - 1;

    /* number of channels in radio... to do this we need to look    */
    /* at all the rx channels and count any in the range of 136-174 */
    chan = 0;
    for ( i = 0; i<16; i++ )
    {
        wid = lookup_widget(app1,rxfreq[i]);
        txt = gtk_entry_get_text((GtkEntry *)wid);
        freq = atof(txt);
        if ( freq >= 136.0 && freq <= 174.0 ) chan++;
    }
    if ( chan < 1 )
        fprintf(stderr,"Warning, no valid channels found\n");
    x2212[0x80] = chan - 1;
    
    /* now lets fill in all the channels */
    for ( i = 0; i<16; i++ )
    {
        /* rx freq lookup off the screen */
        wid = lookup_widget(app1,rxfreq[i]);
        txt = gtk_entry_get_text((GtkEntry *)wid);
        freq = atof(txt);

        /* 3 gross assumptions!!!, timebomb below */
        ref = 13.2e6;
        space = 5000.0;
#define REFCODE 2        

        /* point to the 8 bytes of rx channel data for specified channel */
        chbuff = &x2212[0] + (((i+1) % 16) * 16);
        /* there are 20 packed bits of freq info... lets build the numbers */
        
        /* fill in info for valid rx channels */
        if ( freq >= 136.0 && freq <= 174.0 )
        {
            /* we have a choice of 3 ch spacings when building info */
            /* for a particular rx and tx channel pair              */
            /* as a start, we assume 5khz will do.. should probably */
            /* try all 3 and use the one with the least error       */
            divisor = (((freq + 45.0)*1e6) + (space/2.0)) / space;
            realfreq = ((divisor * space ) - 45e6)/1e6;  /* in mhz */
            if ( fabs(freq-realfreq) > 0.001 )
                fprintf(stderr,"Warning ch %d freq error: %f <> %fmhz\n",
                        i+1,realfreq,freq); 
            chbuff[1] = REFCODE + (( divisor & 0x8000) >> 12);
            chbuff[2] = (divisor & 0x780) >> 7;
            chbuff[3] = (divisor & 0x7800) >> 11;
            chbuff[4] = divisor & 0xf;
            chbuff[5] = ((divisor & 0x30) >> 4) + ((divisor & 0x40) >> 3);

        }
        /* fill in the rx cg code (1st 2 digits of entrybox) */
        wid = lookup_widget(app1,rxcg[i]);
        txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));            
        chbuff[6] = hex2bin(txt[0]);
        chbuff[7] = hex2bin(txt[1]);

        /* now fill in the tx info */
        /* tx freq lookup off the screen */
        wid = lookup_widget(app1,txfreq[i]);
        txt = gtk_entry_get_text((GtkEntry *)wid);
        freq = atof(txt);

        /* point to the 8 bytes of tx channel data for specified channel */
        chbuff = &x2212[0] + (((i+1) % 16) * 16) + 8;
        /* there are 20 packed bits of freq info... lets build the numbers */
        
        /* fill in info for valid tx channels */
        if ( freq >= 136.0 && freq <= 174.0 )
        {
            divisor = ((freq*1e6) + (space/2.0)) / space;
            realfreq = divisor * space /1e6;  /* in mhz */
            if ( fabs(freq-realfreq) > 0.001 )
                fprintf(stderr,"Warning ch %d freq error: %f <> %fmhz\n",
                        i+1,realfreq,freq); 
            chbuff[1] = ( divisor & 0x8000) >> 12;
            chbuff[2] = (divisor & 0x780) >> 7;
            chbuff[3] = (divisor & 0x7800) >> 11;
            chbuff[4] = divisor & 0xf;
            chbuff[5] = ((divisor & 0x30) >> 4) + ((divisor & 0x40) >> 3);
        }
        /* fill in the tx cg code (1st 2 digits of entrybox) */
        wid = lookup_widget(app1,txcg[i]);
        txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));            
        chbuff[6] = hex2bin(txt[0]);
        chbuff[7] = hex2bin(txt[1]);
        
        /* set the bit that enable tx timer for channel if needed */
        wid = lookup_widget(app1,cct[i]);
        if (gtk_toggle_button_get_active((GtkToggleButton *)wid))
            chbuff[1] = chbuff[1] + 0x04;
    }
}



//=====================================================================
// 
// 
//=====================================================================
void
on_write1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *wid;
    int err;
    int i;

    wid = lookup_widget(app1,"appbar1");
   	gnome_appbar_set_status((GnomeAppBar *)wid,"Encoding Rom Image");
    gnome_appbar_set_progress_percentage((GnomeAppBar *)wid,0.0);  
    
   	gnome_appbar_set_status((GnomeAppBar *)wid,"Encoding Rom Image");
    encoderom();

    outb(4 & ~DIRBIT,LPTCTRL);
   	gnome_appbar_set_status((GnomeAppBar *)wid,"Writing Rom Image to Adapter");    
    for (i=0; i<256; i++)
    {
        /* output the prom data */
        outb(x2212[i],LPTDAT);
        usleep(1);
        
        /* latch the prom data */
        outb(5 & ~DIRBIT,LPTCTRL);
        usleep(1);
        gtk_main_iteration_do(FALSE);        
        
        /* next 4 lines not really neded if using 373 latch */
        outb(4 & ~DIRBIT,LPTCTRL);
        usleep(1);
        outb(5 & ~DIRBIT,LPTCTRL);
        usleep(1);

        /* output the prom address */
        outb(i,LPTDAT);
        usleep(1);
        gtk_main_iteration_do(FALSE);        
                
        /* write the data to prom */
        outb(7 & ~DIRBIT,LPTCTRL);
        usleep(1);
        outb(5 & ~DIRBIT,LPTCTRL);
        usleep(1);
        outb(4 & ~DIRBIT,LPTCTRL);
        usleep(1);
        gnome_appbar_set_progress_percentage((GnomeAppBar *)wid,(double)i/255.0);  
        gtk_main_iteration_do(FALSE);
        
    }

    /* verify is done inline here before storing data in eeprom */
    /* to avoid excessive store commands                        */
   	gnome_appbar_set_status((GnomeAppBar *)wid,"Verifying Rom Image with Adapter");
    err = FALSE;
    for (i=0; i<256; i++)
    {
        /* output the address */
        outb(i,LPTDAT);
        usleep(1);
        gtk_main_iteration_do(FALSE);        

        /* strobe read */
        outb(6 & ~DIRBIT,LPTCTRL);
        usleep(1);

        /* compare the data */
        if ( x2212[i] != ((inb(LPTSTAT)/8) & 0x0f))
        {
            fprintf(stderr,"write rom failed: read 0x%02X, should be 0x%02X\n",
                    ((inb(LPTSTAT)/8) & 0x0f),x2212[i]);
            err = TRUE;
        }

        outb(4 & ~DIRBIT,LPTCTRL);
        usleep(1);
        gnome_appbar_set_progress_percentage((GnomeAppBar *)wid,(double)i/255.0);  
        gtk_main_iteration_do(FALSE);
    }    

    if ( err == FALSE )
    {
        /* strobe the ram image into the eeprom image in the x2212 */
        outb(0 & ~DIRBIT,LPTCTRL);
        usleep(1);
        outb(4 & ~DIRBIT,LPTCTRL);
        usleep(1);
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"Writing Complete");        
    }
    else
    {
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"Writing Failed");            
    }    
    gnome_appbar_set_progress_percentage((GnomeAppBar *)wid,0.0);  
}

//=====================================================================
// 
// This reads, rom, and verifies it matches screen contents
//=====================================================================
void
on_verify1_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    unsigned char rom[256];
    int errcnt = 0;
    GtkWidget *wid;    
    int i;
    
    wid = lookup_widget(app1,"appbar1");    
    /* build rom image from screen */
   	gnome_appbar_set_status((GnomeAppBar *)wid,"Verify... building image");        
    encoderom();
    memcpy(rom,x2212,sizeof( rom ));
    
    /* read rom image from adapter */
    /* note. this will overwrite screen with actual rom contents */
    on_read1_activate( NULL, NULL );
    
   	gnome_appbar_set_status((GnomeAppBar *)wid,"Verify... comparing images");
    for ( i=0; i<256; i++ )
    {
        if ( rom[i] != x2212[i] )
        {
            fprintf(stderr,
            "Opps verify failed: x2212[0x%02X]=0x%02X, computed[0x%02X]=0x%02X\n",
                     i,x2212[i] & 0x0f, i, rom[i] & 0x0f);
            errcnt++;
        }
    }                  
    if ( errcnt > 0 )
       	gnome_appbar_set_status((GnomeAppBar *)wid,"Verify... FAILED");
    else
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"Verify... PASSED");    
    
}

//=====================================================================
// 
// 
//=====================================================================
void
on_test_adapter1_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    int i;
    int value;
    int errcnt = 0;
    GtkWidget *wid;
    
    /* DIRBIT forced low. set laptop bidirectional port to o/p */
    outb(4 & ~DIRBIT,LPTCTRL);
    for ( i=0; i<16; i++)
    {
        /* output data */
        outb(i,LPTBASE);
        /* latch data into 74ls373 or 374 */
        outb(5 & ~DIRBIT,LPTCTRL);
        /* next two lines needed if you used 74ls274 */
        outb(4 & ~DIRBIT,LPTCTRL);
        outb(5 & ~DIRBIT,LPTCTRL);
        value = (inb(LPTSTAT)/8) & 0x0f;
        if ( i != value )
        {
            fprintf(stderr,"  Failed... expecting %d read %d\n",i,value);
            errcnt++;
        }
    }
    wid = lookup_widget(app1,"appbar1");
    if ( errcnt == 0 )
    	gnome_appbar_set_status((GnomeAppBar *)wid,"Adapter Working OK");
    else
    	gnome_appbar_set_status((GnomeAppBar *)wid,"Adapter FAILED Loopback Test");     

}


//=====================================================================
// 
// 
//=====================================================================
void
on_app1_realize                        (GtkWidget       *widget,
                                        gpointer         user_data)
{
    /* we need to init all the cg boxes.... doing it here saves a LOT of code */
    /* as there are 32 cg lists all with the same items                       */
    GtkWidget *wid;
    GList *items = NULL;
    int i;
    
    wid = lookup_widget(app1,"appbar1");    
   	gnome_appbar_set_status((GnomeAppBar *)wid,"See GE LBI31360B for reference");        

    items = g_list_append( items, "00 = No CG");
    items = g_list_append( items, "10 = 67.0");
    items = g_list_append( items, "18 = 67.0 ste");
    items = g_list_append( items, "20 = 71.9");
    items = g_list_append( items, "28 = 71.9 ste");
    items = g_list_append( items, "30 = 74.4");
    items = g_list_append( items, "38 = 74.4 ste");
    items = g_list_append( items, "40 = 77.0");
    items = g_list_append( items, "48 = 77.0 ste");
    items = g_list_append( items, "50 = 79.7");
    items = g_list_append( items, "58 = 79.7 ste");
    items = g_list_append( items, "60 = 82.5");
    items = g_list_append( items, "68 = 82.5 ste");
    items = g_list_append( items, "70 = 85.4");
    items = g_list_append( items, "78 = 85.4 ste");
    items = g_list_append( items, "80 = 88.5");
    items = g_list_append( items, "88 = 88.5 ste");
    items = g_list_append( items, "90 = 91.5");
    items = g_list_append( items, "98 = 91.5 ste");
    items = g_list_append( items, "A0 = 94.8");
    items = g_list_append( items, "A8 = 94.8 ste");
    items = g_list_append( items, "12 = 97.4");
    items = g_list_append( items, "1A = 97.4 ste");
    items = g_list_append( items, "B0 = 100.0");
    items = g_list_append( items, "B8 = 100.0 ste");
    items = g_list_append( items, "C0 = 103.5");
    items = g_list_append( items, "C8 = 103.5 ste");
    items = g_list_append( items, "D0 = 107.2");
    items = g_list_append( items, "D8 = 107.2 ste");
    items = g_list_append( items, "E0 = 110.9");
    items = g_list_append( items, "E8 = 110.9 ste");
    items = g_list_append( items, "F0 = 114.8");
    items = g_list_append( items, "F8 = 114.8 ste");
    items = g_list_append( items, "01 = 118.8");
    items = g_list_append( items, "09 = 118.8 ste");
    items = g_list_append( items, "11 = 123.0");
    items = g_list_append( items, "19 = 123.0 ste");
    items = g_list_append( items, "21 = 127.3");
    items = g_list_append( items, "29 = 127.3 ste");
    items = g_list_append( items, "31 = 131.8");
    items = g_list_append( items, "39 = 131.8 ste");
    items = g_list_append( items, "41 = 136.5");
    items = g_list_append( items, "49 = 136.5 ste");
    items = g_list_append( items, "51 = 141.3");
    items = g_list_append( items, "59 = 141.3 ste");
    items = g_list_append( items, "61 = 146.2");
    items = g_list_append( items, "69 = 146.2 ste");
    items = g_list_append( items, "71 = 151.4");
    items = g_list_append( items, "79 = 151.4 ste");
    items = g_list_append( items, "81 = 156.7");
    items = g_list_append( items, "89 = 156.7 ste");
    items = g_list_append( items, "91 = 162.2");
    items = g_list_append( items, "99 = 162.2 ste");
    items = g_list_append( items, "A1 = 167.9");
    items = g_list_append( items, "A9 = 167.9 ste");
    items = g_list_append( items, "B1 = 173.8");
    items = g_list_append( items, "B9 = 173.8 ste");
    items = g_list_append( items, "C1 = 179.9");
    items = g_list_append( items, "C9 = 179.9 ste");
    items = g_list_append( items, "D1 = 186.2");
    items = g_list_append( items, "D9 = 186.2 ste");
    items = g_list_append( items, "E1 = 192.8");
    items = g_list_append( items, "E9 = 192.8 ste");
    items = g_list_append( items, "F1 = 203.5");
    items = g_list_append( items, "F9 = 203.5 ste");
    items = g_list_append( items, "02 = 210.7");
    items = g_list_append( items, "0A = 210.7 ste");

    for ( i=0; i<16; i++ )
    {
        wid = lookup_widget(app1,rxcg[i]);
        gtk_combo_set_popdown_strings(GTK_COMBO(wid),items);
        wid = lookup_widget(app1,txcg[i]);
        gtk_combo_set_popdown_strings(GTK_COMBO(wid),items);
    }

    /* to read and write to the parallel port, we need it open as root */
    fd0 = open("/dev/lp0", O_RDWR | O_NONBLOCK);
    if (fd0 >= 0) 
    {
        int stat = ioperm(LPTDAT, 3, 1);
    	if ( stat == -1 )
	        perror("ioperm failed ");
    }
    else
    {
        fprintf(stderr,"\n\n*** open on /dev/lp0 failed...Do you have permission?\n");
        fprintf(stderr,"*** you may need to 'sudo modprobe parport_pc'\n");
        fprintf(stderr,"*** 'sudo ge-x2212-2'\n");
    	perror("/dev/lp0");
        done();
    }
//  g_print("Using port: 0x%x \n", LPTDAT);

    /* DIRBIT forced low. set laptop bidirectional port to o/p */
    outb(4 & ~DIRBIT,LPTCTRL);


//    g_print(" end of app1_realize\n");

}


//=====================================================================
// 
// Show the user about who did what
//=====================================================================
void
on_about2_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *about2;
    about2 = create_about2();
    gtk_widget_show (about2);
}

//=====================================================================
// 
// Dont forget to add this so app closes properly on "X"
//=====================================================================
gboolean
on_app1_delete_event                   (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    done();
	gtk_main_quit();    
    return FALSE;
}
//=====================================================================
// 
// 
//=====================================================================
/* The file selection widget and the string to store the chosen filename */

void export_filename (GtkWidget *widget, gpointer user_data) 
{
    FILE *fp;
    int i;
    GtkWidget *wid;
    GtkWidget *file_selector = (GtkWidget *)user_data;
    G_CONST_RETURN gchar *txt;
        
    /* make sure rom image is up to date */
    encoderom();
    /* make sure screen is up to date (mostly active channels) */
    decoderom();
    
    wid = lookup_widget(app1,"appbar1");

    selected_filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_selector));
    
    fp = fopen(selected_filename,"wt");
    if ( fp == NULL )
    {
        wid = lookup_widget(app1,"appbar1");
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"ERROR - unable to open file");
        return;
    }
    else
    {
        /* write a 16 line txt file with freq info for possible printing */
        fprintf(fp,"ge-x2212-2: GE Phoenix SX 16ch Scanning Model Frequency List\n");
        fprintf(fp,"by Lawrence Glaister VE7IT, (c)2004\n\n");
        
        wid = lookup_widget(app1,"numch");
        fprintf(fp,"Number of Active Channels: %d\n",(x2212[0x80]&0x0f) + 1);

        wid = lookup_widget(app1,"scanoption");
        txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));
        fprintf(fp,"Scan Option: %s\n",txt);

        wid = lookup_widget(app1,"txtimer");
        txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));
        fprintf(fp,"Radio Wide Tx Timer: %s\n",txt);

        wid = lookup_widget(app1,"fixch");
        txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));
        fprintf(fp,"Fixed Priority Channel: %s\n\n",txt);

        fprintf(fp,"Rx Freq   Rx CG             Tx Freq   Tx CG            CCT\n");
        fprintf(fp,"==========================================================\n");                
        for ( i=0; i<16; i++)
        {
            wid = lookup_widget(app1,rxfreq[i]);
            txt = gtk_entry_get_text((GtkEntry *)wid);
            fprintf(fp,"%-10s",txt);

            wid = lookup_widget(app1,rxcg[i]);
            txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));
            fprintf(fp,"%-18s",txt);
            
            wid = lookup_widget(app1,txfreq[i]);
            txt = gtk_entry_get_text((GtkEntry *)wid);
            fprintf(fp,"%-10s",txt);

            wid = lookup_widget(app1,txcg[i]);
            txt = gtk_entry_get_text((GtkEntry *)(((GtkCombo *)wid)->entry));
            fprintf(fp,"%-18s",txt);

            wid = lookup_widget(app1,cct[i]);
            if (gtk_toggle_button_get_active((GtkToggleButton *)wid))
                fprintf(fp," Y\n");
            else
                fprintf(fp," N\n");            
        }
        fclose(fp);
        wid = lookup_widget(app1,"appbar1");
   	    gnome_appbar_set_status((GnomeAppBar *)wid,"Exported frequency list");
    }
}

//=====================================================================
// 
// Writes a printable freq list from screen data
//
//=====================================================================
void
on_export1_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *file_selector;
    file_selector = gtk_file_selection_new("Text Filename To Save...");
    gtk_file_selection_complete(GTK_FILE_SELECTION(file_selector),"*.txt");

    g_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->ok_button),
                     "clicked",
                     G_CALLBACK (export_filename),
                     (gpointer) file_selector);
   			   
   /* Ensure that the dialog box is destroyed when the user clicks a button. */
   
   g_signal_connect_swapped (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->ok_button),
                             "clicked",
                             G_CALLBACK (gtk_widget_destroy), 
                             (gpointer) file_selector); 

   g_signal_connect_swapped (GTK_OBJECT (GTK_FILE_SELECTION (file_selector)->cancel_button),
                             "clicked",
                             G_CALLBACK (gtk_widget_destroy),
                             (gpointer) file_selector); 
   
   /* Display that dialog */
   gtk_widget_show (file_selector);

}

